#!/usr/bin/env python

try:
    from functools import cached_property
except ImportError:
    from backports.cached_property import cached_property
import os
import subprocess
import configparser
from pathlib import Path
from .clickup import ClickUpClient
from .exceptions import GitException, ConfigException


__all__ = ["Config", "Git", "Workflow"]

CLICKUP_TOKEN = "CLICKUP_TOKEN"
DEFAULT_GIT_BASE_BRANCH = "main"


class Git:
    def __init__(self, config):
        self.config = config

    def run(self, *args, git_dir=None):
        git_dir = git_dir or (self.config and self.config.git_dir) or None
        if git_dir:
            args = ["git", "-C", git_dir] + list(args)
        else:
            args = ["git"] + list(args)
        completed_process = subprocess.run(args, capture_output=True)
        if completed_process.returncode != 0:
            raise GitException(completed_process.stderr.decode("utf-8"))
        return completed_process.stdout.strip().decode("utf-8")

    def run_gh(self, *args, git_dir=None):
        git_dir = git_dir or (self.config and self.config.git_dir) or None
        args = ["gh"] + list(args)
        completed_process = subprocess.run(args, capture_output=True, cwd=git_dir)
        if completed_process.returncode != 0:
            raise GitException(completed_process.stderr.decode("utf-8"))
        return completed_process.stdout.strip().decode("utf-8")

    def get_current_branch(self):
        "Get current branch"
        return self.run("rev-parse", "--abbrev-ref", "HEAD")

    def get_toplevel(self, git_dir=None):
        "Get the path of the top-level directory of the working tree"
        return Path(self.run("rev-parse", "--show-toplevel", git_dir=git_dir))

    def get_remote_url(self):
        "Get remote url"
        try:
            return self.run("remote", "get-url", "origin")
        except GitException:
            return None

    def get_github_url(self, branch_name):
        " Get link to github (if origin is github)"
        url = self.get_remote_url()
        if not url or "github.com" not in url:
            return None
        if url.startswith("git@github.com:"):
            url = url.replace("git@github.com:", "https://github.com/")
        url = url[:-4]
        if branch_name:
            url = url + "/tree/" + branch_name
        return url


class Config:

    default_clickup_token = None
    default_clickup_team_id = None
    git_dir = None
    git_base_branch = DEFAULT_GIT_BASE_BRANCH

    def __init__(self, base_path=None):
        # Load credentials
        self.load_credentials()
        # Get project config
        self.load_config(base_path)

    def load_credentials(self):
        "Load credentials - load token from CLICKUP_TOKEN environment variable or ~/.alkemy_workflow/credentials file"
        credentials_path = self.get_credentials_path()
        if CLICKUP_TOKEN in os.environ:
            self.default_clickup_token = os.environ[CLICKUP_TOKEN]
        elif credentials_path.exists():
            cp = configparser.ConfigParser()
            cp.read(credentials_path)
            self.retrieve_config(cp, "default", "clickup_token")
            self.retrieve_config(cp, "default", "clickup_team_id")
        if not self.default_clickup_token:
            raise ConfigException(
                f"Please set the ClickUp token in {credentials_path} file or in the {CLICKUP_TOKEN} environment variable"
            )

    def load_config(self, base_path):
        "Get project config - load alkemy_workflow.ini file from project root folder"
        git = Git(self)
        self.git_dir = git.get_toplevel(base_path)
        self.config_path = self.git_dir / "alkemy_workflow.ini"
        if self.config_path.exists():
            cp = configparser.ConfigParser()
            cp.read(self.config_path)
            self.retrieve_config(cp, "git", "base_branch")

    def retrieve_config(self, cp, section, key):
        try:
            setattr(self, section + "_" + key, cp[section][key])
        except KeyError:
            pass

    def write_config(self, base_branch=None):
        "Write project config file alkemy_workflow.ini"
        cp = configparser.ConfigParser()
        if self.config_path.exists():
            cp.read(self.config_path)
        if not cp.has_section("git"):
            cp["git"] = {}
        if base_branch:
            cp.set("git", "base_branch", base_branch)
        elif not cp.has_option("git", "base_branch"):
            git = Git(self)
            cp.set("git", "base_branch", git.get_current_branch())
        with open(self.config_path, "w") as f:
            cp.write(f)

    @classmethod
    def get_credentials_path(cls):
        "Get credential file path"
        home_dir = Path.home()
        return home_dir / ".alkemy_workflow" / "credentials"

    @classmethod
    def write_credentials(cls, token):
        "Write the token in the credentials file"
        if not token.startswith("pk_"):
            raise ConfigException("Tokens will always begin with pk_")
        credentials_path = cls.get_credentials_path()
        cp = configparser.ConfigParser()
        if credentials_path.exists():
            cp.read(credentials_path)
        if not cp.has_section("default"):
            cp["default"] = {}
        cp.set("default", "clickup_token", token)
        os.makedirs(credentials_path.parent, exist_ok=True)
        with open(credentials_path, "w") as f:
            cp.write(f)


class Workflow:
    def __init__(self, base_path=None):
        self.base_path = base_path

    @cached_property
    def config(self):
        return Config(base_path=self.base_path)

    @cached_property
    def git(self):
        return Git(self.config)

    @cached_property
    def client(self):
        return ClickUpClient(self.config)



#!/usr/bin/env python

from pathlib import Path
from .exceptions import InvalidOption, CommandNotFound, ShowHelp
from .utils import Workflow

__all__ = [
    "cmd",
    "cmds",
    "lookup_cmd",
    "parse_args",
    "parse_usage",
]

cmds = []


def cmd(usage="", defaults=None):
    "Command decorator"

    def wrap(f):
        def wrapped_f(argv):
            kargs = parse_args(argv, usage, defaults)
            wf = Workflow(kargs.get("C"))
            f(kargs, wf)

        wrapped_f.__name__ = f.__name__
        wrapped_f.usage = usage
        wrapped_f.__doc__ = getattr(f, "__doc__", None)
        wrapped_f.cmd = f.__name__[4:] if f.__name__.startswith("cmd_") else f.__name__
        cmds.append(wrapped_f)
        return wrapped_f

    return wrap


def parse_args(args, usage, defaults=None):
    "Parse command lines arguments"
    required, optionals, arguments = parse_usage(usage)
    result = []
    kargs = dict(defaults or {})
    kargs["positional"] = result
    opt = None
    args = list(args)
    if args:
        kargs["exe"] = Path(args.pop(0)).name  # args[0] is the executable file name
        if kargs["exe"] == "__main__.py":
            kargs["exe"] = "aw"
    if args:
        kargs["cmd"] = args.pop(0)  # args[0] is the command
    while args:
        arg = args.pop(0)
        if opt is not None:  # arguments value
            value = arg
            kargs[opt] = value
            opt = None
        elif arg in ("--help", "-h", "-?"):
            raise ShowHelp()
        elif required:  # required positional argument
            result.append(arg)
            required.pop(0)
        elif optionals and arg[0] != "-":  # optional positional argument
            result.append(arg)
            optionals.pop(0)
        elif "=" in arg:  # argument --key=value
            (opt, next_arg) = arg.split("=", 1)
            if not opt.startswith("--"):
                raise InvalidOption(f"invalid option: {arg}")
            opt = opt[2:]
            args.insert(0, next_arg)
        else:
            if not arg.startswith("-"):
                raise InvalidOption(f"invalid option: {arg}")
            t = arg.lstrip("-")
            if t in ("help", "h", "?"):
                raise ShowHelp()
            if arguments.get(t) == bool:  # boolean arg
                kargs["op_" + t] = True
            else:
                opt = t
    if opt is not None:
        raise InvalidOption(f"missing value for {opt}")
    if required:  # check missing required values
        raise InvalidOption("missing {}".format(required.pop(0)))
    while optionals:  # add missing optional values
        result.append(None)
        optionals.pop()
    return kargs


def parse_usage(usage):
    "Parse usage help line"
    usage = usage.split("\n")[0].split()
    required = []
    optionals = []
    arguments = {}
    while usage:
        item = usage.pop(0)
        if not item.startswith("["):
            required.append(item)
        else:
            item = item.strip("[]")
            if item[0] != "-":
                optionals.append(item)
            else:
                item = item.lstrip("-")
                if "=" in item:
                    item = item.split("=")[0]
                    arguments[item] = str
                else:
                    arguments[item] = bool
    return required, optionals, arguments


def lookup_cmd(cmd):
    for f in cmds:
        if cmd == getattr(f, "cmd"):
            return f
    raise CommandNotFound('Invalid command "{}"; use "help" for a list.'.format(cmd))



#!/usr/bin/env python
import os
from pathlib import Path
from setuptools import setup, find_packages

HERE = Path(__file__).parent

install_requires = [line.rstrip() for line in open(HERE / "requirements.txt")]

VERSION_FILE = HERE / "alkemy_workflow" / "VERSION"

with open(VERSION_FILE) as f:
    version = f.read().strip()

setup(
    name="alkemy_workflow",
    version=version,
    description="Alkemy Workflow",
    long_description="Alkemy Git Workflow",
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="git workflow",
    url="https://github.com/DLBD-Department/alkemy-worflow",
    license="MIT",
    packages=find_packages(exclude=["ez_setup", "examples"]),
    include_package_data=True,
    zip_safe=True,
    install_requires=install_requires,
    entry_points={
        "console_scripts": [
            "aw=alkemy_workflow.cli:main",
        ],
    },
    test_suite="tests",
    tests_require=["pytest"],
)



#!/usr/bin/env python

__all__ = [
    "GenericException",
    "GenericWarning",
    "TaskNotFound",
    "SpaceNotFound",
    "FolderNotFound",
    "ListNotFound",
    "StatusNotFound",
    "InvalidOption",
    "CommandNotFound",
    "ShowHelp",
    "GitException",
    "ConfigException",
    "ClickUpException",
]


class GenericException(Exception):
    "Generic exception"


class GenericWarning(GenericException):
    "Generic warning (exit with 0 status)"


class TaskNotFound(GenericException):
    "Task not found"


class SpaceNotFound(GenericException):
    "Space not found"


class FolderNotFound(GenericException):
    "Folder not found"


class ListNotFound(GenericException):
    "List not found"


class StatusNotFound(GenericException):
    "Status not found"


class InvalidOption(GenericException):
    "Invalid option (command line argument)"


class CommandNotFound(GenericException):
    "Command not found"


class ShowHelp(GenericException):
    "Show help message"


class GitException(GenericException):
    "Git exception"


class ConfigException(GenericException):
    "Missing configuration exception"


class ClickUpException(GenericException):
    "ClickUp exception"



#!/usr/bin/env python

import sys
from datetime import datetime
from pathlib import Path
from .exceptions import (
    GenericWarning,
    GenericException,
    GitException,
    InvalidOption,
    ShowHelp,
)
from .cmds import cmd, cmds, lookup_cmd
from .utils import Config

EXIT_SUCCESS = 0
EXIT_FAILURE = 1
EXIT_PARSER_ERROR = 2

__all__ = ["main"]


def pairwise(iterable):
    return zip(iterable, iterable[1:] + [None])


def prepare_tree(items, enabled=True):
    if not enabled:
        for item in items:
            item["tree"] = str(item["id"])
            yield item
    else:
        level = 0
        for item, next_item in pairwise(items):
            is_last = (next_item is None) or (item["type"] != next_item["type"])
            if level == 0:
                graph = ""
            elif not is_last:
                graph = "├─ "
            else:
                graph = "└─ "
            item["tree"] = f"{' ' * level}{graph}{item.id}"
            if is_last:
                level = level + 1
            yield item


@cmd("[-C=cwd] [--filter=filter] [--noheaders]", defaults={"op_noheaders": False})
def cmd_spaces(kargs, wf):
    """
    List spaces
    Example: spaces --filter 'test*' --noheaders
    """
    header = not kargs["op_noheaders"]
    result = wf.client.query(filter_type="Space", filter_name=kargs.get("filter"))
    fmt = "{id:15} {name:40}"
    if header:
        print(fmt.format(id="Id", name="Space"))
        print("-" * 70)
    for item in result:
        print(fmt.format(**item))


@cmd(
    "[-C=cwd] [--space=space] [--filter=filter] [--noheaders]",
    defaults={"op_noheaders": False, "space": None},
)
def cmd_folders(kargs, wf):
    """
    List folders from a space
    Example: folders --space 'test' --filter 'abc*' --noheaders
    """
    header = not kargs["op_noheaders"]
    if not kargs["space"]:
        raise InvalidOption("Missing option space")
    result = wf.client.query(
        space=kargs["space"], filter_type="Folder", filter_name=kargs.get("filter")
    )
    fmt = "{id:15} {name:40}"
    if header:
        print(fmt.format(id="Id", name="Folders"))
        print("-" * 70)
    for item in result:
        print(fmt.format(**item))


@cmd(
    "[-C=cwd] [--space=space] [--folder=folder] [--filter=filter] [--noheaders]",
    defaults={"op_noheaders": False, "space": None, "folder": None},
)
def cmd_lists(kargs, wf):
    """
    List lists from a space or folder
    Example: lists --space 'test' --filter 'abc*' --noheaders
    """
    header = not kargs["op_noheaders"]
    if not kargs["space"] and not kargs["folder"]:
        raise InvalidOption("Missing option space/folder")
    result = wf.client.query(
        space=kargs["space"],
        folder=kargs["folder"],
        filter_type="List",
        filter_name=kargs.get("filter"),
    )
    fmt = "{id:15} {name:40}"
    if header:
        print(fmt.format(id="Id", name="Lists"))
        print("-" * 70)
    for item in result:
        print(fmt.format(**item))


@cmd(
    "[-C=cwd] [--task_id=task_id] [--space=space] [--folder=folder] [--list=list] [--task=task] [--filter=filter] [--noheaders] [--long]",
    defaults={"op_noheaders": False, "op_long": False},
)
def cmd_tasks(kargs, wf):
    """
    List tasks from a list or subtask
    """
    header = not kargs["op_noheaders"]
    if not kargs.get("list") and not kargs.get("task"):
        raise InvalidOption("Missing option list or task")
    result = wf.client.query(
        space=kargs.get("space"),
        folder=kargs.get("folder"),
        lst=kargs.get("list"),
        task=kargs.get("task"),
        filter_name=kargs.get("filter"),
    )
    fmt = "{label:15} {id:15} {name:40}"
    if header:
        print(fmt.format(id="Id", label="Status", name="Title"))
        print("-" * 70)
    for item in result:
        print(fmt.format(**item))


@cmd(
    "[-C=cwd] [--space=space] [--folder=folder] [--list=list] [--task=task] [--filter=filter] [--noheaders] [--nohierarchy]",
    defaults={"op_noheaders": False, "op_nohierarchy": False},
)
def cmd_ls(kargs, wf):
    """
    List spaces --> folders --> lists --> tasks --> subtasks
    Example: ls --space 'test' --noheaders
    """
    header = not kargs["op_noheaders"]
    hierarchy = not kargs["op_nohierarchy"]
    result = wf.client.query(
        space=kargs.get("space"),
        folder=kargs.get("folder"),
        lst=kargs.get("list"),
        task=kargs.get("task"),
        filter_name=kargs.get("filter"),
        hierarchy=hierarchy,
    )
    fmt = "{tree:20} {label:15} {name:40}"
    if header:
        print(fmt.format(label="Kind/Status", tree="Id", name="Title"))
        print("-" * 70)
    for item in prepare_tree(result, enabled=hierarchy):
        print(fmt.format(**item))


@cmd("[-C=cwd] task_id")
def cmd_branch(kargs, wf):
    """
    Open a task and create a new git branch
    Example: branch '#12abcd45'
    """
    task_id = kargs["positional"][0]
    task = wf.client.get_task_by_id(task_id)
    branch_already_exists = False
    # Create a new branch a switch to it
    wf.git.run("checkout", wf.config.git_base_branch)
    try:
        wf.git.run("checkout", "-b", task.branch_name)
    except GitException:
        branch_already_exists = True
        wf.git.run("checkout", task.branch_name)
    try:
        wf.git.run("push", "--set-upstream", "origin", task.branch_name)
    except GitException:
        pass
    # Update the task
    task_update = {}  # task fields to be updated
    # Start date
    if not task.get("start_date"):
        task_update["start_date"] = int(datetime.utcnow().timestamp() * 1000)
    # Task assignee
    current_user = wf.client.get_user()
    if current_user["id"] not in [x["id"] for x in task["assignees"]]:
        task_update["assignees"] = {
            "add": [ current_user["id"] ]
        }
    # Status
    if task.get("status", {}).get("status") == "to do":
        task_update["status"] = "in_progress"
    # # Open (start) the task
    task.update_task(**task_update)
    # Post a comment
    if not branch_already_exists:
        github_url = wf.git.get_github_url(task.branch_name)
        if github_url:
            comment = f"Branch [{task.branch_name}]({github_url})"
        else:
            comment = f"Branch {task.branch_name}"
        task.post_task_comment(comment)


@cmd("[-C=cwd] [-m=message]")
def cmd_commit(kargs, wf):
    """
    Create a new commit an the current feature branch
    Example: commit
    """
    # Get task_id by branch name
    current_branch = wf.git.get_current_branch()
    task_id = current_branch.split("-")[0]
    if task_id == wf.config.git_base_branch:
        raise GenericException(
            f"Please commit from feature branches, not {wf.config.git_base_branch}"
        )
    task = wf.client.get_task_by_id(task_id)
    # Prepare the commit message
    message = f"[{task['id']}] {task['name']}"
    if kargs.get("m"):
        message = message + " - " + kargs["m"]
    # Commit
    print(wf.git.run("commit", "-m", message))


@cmd("[-C=cwd]")
def cmd_pr(kargs, wf):
    """
    Push local commits to the remote branch and create a pull request on GitHub
    Example: pr
    """
    wf.git.run("push")
    wf.git.run_gh("pr", "create", "--fill")


@cmd("[-C=cwd]")
def cmd_status(kargs, wf):
    """
    Status
    Example: status
    """
    # Get task_id by branch name
    current_branch = wf.git.get_current_branch()
    task_id = current_branch.split("-")[0]
    # Print task
    if task_id != wf.config.git_base_branch:
        task = wf.client.get_task_by_id(task_id)
        fmt = """\
Task id: {task.task_id}
Status:  {task.status}
Space:   {task.space}
Folder:  {task.folder}
List:    {task.list}
Title:   {task.title}
"""
        print(fmt.format(task=task))
    print(wf.git.run("status"))


@cmd("[--clickup-token=token]")
def cmd_configure(kargs, wf):
    """
    Configures credentials (Clickup API token).
    Example: configure
    """
    prompt = "ClickUP API token: "
    token = kargs.get("clickup-token") or input(prompt).strip()
    if token:
        Config.write_credentials(token)
    wf.client.get_user()


@cmd("[-C=cwd] [--base-branch=git-base-branch]")
def cmd_init(kargs, wf):
    """
    Write alkemy_workflow.ini config file into the project root folder
    Example: init --base-branch=main
    """
    wf.config.write_config(base_branch=kargs.get("base-branch"))


@cmd("[command]")
def cmd_help(kargs, wf):
    """
    Display information about commands.
    """
    try:
        command = kargs["positional"][0]
    except KeyError:
        command = None
    if command:
        f = lookup_cmd(command)
        usage_text = getattr(f, "usage", "")
        help_text = (getattr(f, "__doc__", "") or "").rstrip("\n\t ")
        print(f"usage: {kargs['exe']} {f.cmd} {usage_text} {help_text}")
    else:
        print(f"usage: {kargs['exe']} <command> [parameters]")
        print("")
        print("Commands:")
        # for f in sorted(cmds, key=lambda x: getattr(x, "cmd")):
        for f in cmds:
            usage_text = getattr(f, "usage", "")
            help_text = (getattr(f, "__doc__", "") or "").rstrip("\n\t ")
            print(f" {f.cmd} {usage_text} {help_text}")
            print("")


def main(argv=None):
    if argv is None:
        argv = sys.argv
    if len(argv) < 2:
        cmd_help(argv)
        return EXIT_PARSER_ERROR
    exe = Path(argv[0]).name
    try:
        f = lookup_cmd(argv[1])
        f(argv)
        return EXIT_SUCCESS
    except ShowHelp:
        cmd_help([exe, "help", argv[1]])
        return EXIT_SUCCESS
    except GenericWarning as ex:
        print(f"{exe}: {ex}")
        return EXIT_SUCCESS
    except GenericException as ex:
        print(f"{exe}: {ex}")
        return EXIT_FAILURE


if __name__ == "__main__":
    sys.exit(main(sys.argv))
